# Copyright (c) 2010-2016 Bo Lin
# Copyright (c) 2010-2016 Yanhong Annie Liu
# Copyright (c) 2010-2016 Stony Brook University
# Copyright (c) 2010-2016 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import ast
import sys
from da.compiler.dast import *

BOOLOP_SYMBOLS = {
    AndOp:        '∧',
    OrOp:         '∨',
    NotOp:        '¬'
}

QUANTOP_SYMBOLS = {
    ExistentialOp:  '∃',
    UniversalOp:    '∀'
}

BINOP_SYMBOLS = {
    AddOp:        '+',
    SubOp:        '-',
    MultOp:       '*',
    DivOp:        '/',
    FloorDivOp:   '//',
    ModOp:        '%',
    LShiftOp:     '<<',
    RShiftOp:     '>>',
    BitOrOp:      '|',
    BitAndOp:     '&',
    BitXorOp:     '^',
    PowOp:        '**'
}

CMPOP_SYMBOLS = {
    EqOp:         '==',
    GtOp:         '>',
    GtEOp:        '>=',
    InOp:         '∈',
    IsOp:         'is',
    IsNotOp:      'is not',
    LtOp:         '<',
    LtEOp:        '<=',
    NotEqOp:      '≠',
    NotInOp:      '∉'
}

UNARYOP_SYMBOLS = {
    InvertOp:     '~',
    NotOp:        'not',
    UAddOp:       '+',
    USubOp:       '-'
}

EVENT_TYPES = {
    ReceivedEvent: 'receive',
    SentEvent:     'send'
}

# Large float and imaginary literals get turned into infinities in the AST.
# We unparse those infinities to INFSTR.
INFSTR = "1e" + repr(sys.float_info.max_10_exp + 1)

def to_pseudo(tree):
    import io
    textbuf = io.StringIO(newline='')
    DastUnparser(tree, textbuf)
    return textbuf.getvalue()

def interleave(inter, f, seq):
    """Call f on each item in seq, calling inter() in between.
    """
    seq = iter(seq)
    try:
        f(next(seq))
    except StopIteration:
        pass
    else:
        for x in seq:
            inter()
            f(x)

class DastUnparser:
    """Methods in this class recursively traverse an AST and
    output source code for the abstract syntax; original formatting
    is disregarded. """

    def __init__(self, tree, file=sys.stdout, indent_width=4):
        """Unparser(tree, file=sys.stdout) -> None.
         Print the source for tree to file."""
        self.f = file
        self.counter = 0
        self._indent = 0
        self._indent_width = indent_width
        self.dispatch(tree)
        print("", file=self.f)
        self.f.flush()

    def fill(self, text = ""):
        "Indent a piece of text, according to the current indentation level"
        text = ("\n"+" "*self._indent_width*self._indent + text)
        self.f.write(text)
        self.counter += len(text)

    def write(self, text):
        "Append a piece of text to the current line."
        self.f.write(text)
        self.counter += len(text)

    def enter(self):
        "Print ':', and increase the indentation."
        self.write(":")
        self._indent += 1

    def leave(self):
        "Decrease the indentation level."
        self._indent -= 1

    def label(self, label):
        "Print 'label' followed by a ':', indented one space from the previous level."
        offset = (self._indent_width * (self._indent - 1)) + 1
        self.write("\n" + " " * offset + label + ':')

    def dispatch(self, tree):
        "Dispatcher function, dispatching tree type T to method _T."
        if isinstance(tree, list):
            for t in tree:
                self.dispatch(t)
            return
        if isinstance(tree, Statement) and tree.label:
            self.label(tree.label)
        meth = getattr(self, "_"+tree.__class__.__name__)
        meth(tree)


    ############### Unparsing methods ######################
    # There should be one method per concrete grammar type #
    # Constructors should be grouped by sum type. Ideally, #
    # this would follow the order in the grammar, but      #
    # currently doesn't.                                   #
    ########################################################

    def _Program(self, tree):
        # for stmt in tree.processes:
        #     self.dispatch(stmt)
        # if tree.entry_point:
        #     self.dispatch(tree.entry_point)
        self.dispatch(tree.body)

    # stmt
    def _SimpleStmt(self, tree):
        self.fill()
        self.dispatch(tree.expr)

    def _ImportStmt(self, t):
        self.fill("import ")
        interleave(lambda: self.write(", "), self.dispatch, t.items)

    def _ImportFromStmt(self, t):
        self.fill("from ")
        self.write("." * t.level)
        if t.module:
            self.write(t.module)
        self.write(" import ")
        interleave(lambda: self.write(", "), self.dispatch, t.items)

    def _AssignmentStmt(self, t):
        self.fill()
        for target in t.targets:
            self.dispatch(target)
            self.write(" = ")
        self.dispatch(t.value)

    def _OpAssignmentStmt(self, t):
        self.fill()
        self.dispatch(t.target)
        self.write(" " + BINOP_SYMBOLS[t.operator] + "= ")
        self.dispatch(t.value)

    def _ReturnStmt(self, t):
        self.fill("return")
        if t.value:
            self.write(" ")
            self.dispatch(t.value)

    def _NoopStmt(self, t):
        self.fill("pass")

    def _BreakStmt(self, t):
        self.fill("break")

    def _ContinueStmt(self, t):
        self.fill("continue")

    def _DeleteStmt(self, t):
        self.fill("del ")
        interleave(lambda: self.write(", "), self.dispatch, t.targets)

    def _AssertStmt(self, t):
        self.fill("assert ")
        self.dispatch(t.expr)
        if t.msg:
            self.write(", ")
            self.dispatch(t.msg)

    def _GlobalStmt(self, t):
        self.fill("global ")
        interleave(lambda: self.write(", "), self.write, t.names)

    def _NonlocalStmt(self, t):
        self.fill("nonlocal ")
        interleave(lambda: self.write(", "), self.write, t.names)

    def _AwaitStmt(self, t):
        self.fill("await")
        self._do_await_branches(t)

    def _LoopingAwaitStmt(self, t):
        self.fill("awhile")
        self._do_await_branches(t)

    def _do_await_branches(self, t):
        if len(t.branches) == 1 and not t.branches[0].body and not t.orelse:
            # single-line await
            self.write(" ")
            self.dispatch(t.branches[0].condition)
        if t.timeout:
            self.write(" until ")
            self.dispatch(t.timeout)
        if len(t.branches) > 1 or t.branches[0].body or t.orelse:
            self.enter()
            for b in t.branches:
                self.fill("if ")
                self.dispatch(b)
            if t.orelse:
                self.fill("if timeout")
                self.enter()
                self.dispatch(t.orelse)
                self.leave()
            self.leave()


    def _Branch(self, t):
        self.dispatch(t.condition)
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _SendStmt(self, t):
        self.fill("send ")
        self.dispatch(t.message)
        self.write(" to ")
        self.dispatch(t.target)

    def _OutputStmt(self, t):
        self.fill("output ")
        interleave(lambda: self.write(', '), self.dispatch, t.message)
        if t.level:
            self.write(" at level ")
            self.dispatch(t.level)

    def _ResetStmt(self, t):
        self.fill("reset ")
        self.dispatch(t.expr)

    def _YieldStmt(self, t):
        self.write("(")
        self.write("yield")
        if t.value:
            self.write(" ")
            self.dispatch(t.value)
        self.write(")")

    def _YieldFrom(self, t):
        self.write("(")
        self.write("yield from")
        if t.value:
            self.write(" ")
            self.dispatch(t.value)
        self.write(")")

    def _RaiseStmt(self, t):
        self.fill("raise")
        if not t.expr:
            assert not t.cause
            return
        self.write(" ")
        self.dispatch(t.expr)
        if t.cause:
            self.write(" from ")
            self.dispatch(t.cause)

    def _PassStmt(self, t):
        self.fill("pass")

    def _TryStmt(self, t):
        self.fill("try")
        self.enter()
        self.dispatch(t.body)
        self.leave()
        for ex in t.excepthandlers:
            self.dispatch(ex)
        if t.elsebody:
            self.fill("else")
            self.enter()
            self.dispatch(t.elsebody)
            self.leave()
        if t.finalbody:
            self.fill("finally")
            self.enter()
            self.dispatch(t.finalbody)
            self.leave()

    def _ExceptHandler(self, t):
        self.fill("except")
        if t.type:
            self.write(" ")
            self.dispatch(t.type)
        if t.name:
            self.write(" as ")
            self.write(t.name)
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _ClassStmt(self, t):
        self.write("\n")
        for deco in t.decorators:
            self.fill("@")
            self.dispatch(deco)
        self.fill("class " + t.name)
        self.write("(")
        comma = False
        for e in t.bases:
            if comma: self.write(", ")
            else: comma = True
            self.dispatch(e)
        for e in t.keywords:
            if comma: self.write(", ")
            else: comma = True
            self.dispatch(e)
        if t.starargs:
            if comma: self.write(", ")
            else: comma = True
            self.write("*")
            self.dispatch(t.starargs)
        if t.kwargs:
            if comma: self.write(", ")
            else: comma = True
            self.write("**")
            self.dispatch(t.kwargs)
        self.write(")")

        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _Process(self, t):
        self.write("\n")
        for deco in t.decorators:
            self.fill("@")
            self.dispatch(deco)
        self.fill("process " + t.name)
        if t.bases:
            self.write(" extends ")
            comma = False
            for e in t.bases:
                if comma: self.write(", ")
                else: comma = True
                self.dispatch(e)
        self.enter()
        self.label("#PARAMS")
        self.write(" ")
        self.dispatch(t.args)
        self.write("\n")
        self.label('#STATES')
        self.write(" ")
        for name in t.ordered_local_names:
            self.write(name)
            self.write(" ")
        self.write("\n")
        if t.configurations:
            self.label("#CONFIG")
            self.write("\n")
            for key, value in t.configurations:
                self.fill(key + " : ")
                self.dispatch(value)
        if t.setup:
            self.label("#INIT")
            self.dispatch(t.setup.body)
            self.write("\n")
        if t.entry_point:
            self.label("#ENTRY_POINT")
            self.dispatch(t.entry_point)
            self.write("\n")
        if t.events:
            self.label("#EVENTS")
            self.dispatch(t.event_handlers)
            self.write("\n")
        if t.methods:
            self.label("#METHODS")
            self.dispatch(t.methods)
        self.leave()

    def _Function(self, t):
        self.write("\n")
        for deco in t.decorators:
            self.fill("@")
            self.dispatch(deco)
        self.fill("def "+t.name + "(")
        self.dispatch(t.args)
        self.write(")")
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _EventHandler(self, t):
        self.write("\n")
        self.fill("on ")
        first = True
        for evt in t.events:
            if first: first = False
            else: self.write(" or ")
            self.write(EVENT_TYPES[evt.type] + " ")
            self.dispatch(evt)
        if t.labels:
            self.write("at labels ")
            self.write(t.labels)
        if t.notlabels:
            self.write("but not labels ")
            self.write(t.notlabels)
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _ForStmt(self, t):
        self.fill("for ")
        self.dispatch(t.domain)
        self.enter()
        self.dispatch(t.body)
        self.leave()
        if t.elsebody:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave()

    def _IfStmt(self, t):
        self.fill("if ")
        self.dispatch(t.condition)
        self.enter()
        self.dispatch(t.body)
        self.leave()
        # collapse nested ifs into equivalent elifs.
        while (t.elsebody and len(t.elsebody) == 1 and
               isinstance(t.elsebody[0], IfStmt)):
            t = t.elsebody[0]
            self.fill("elif ")
            self.dispatch(t.condition)
            self.enter()
            self.dispatch(t.body)
            self.leave()
        # final else
        if t.elsebody:
            self.fill("else")
            self.enter()
            self.dispatch(t.elsebody)
            self.leave()

    def _WhileStmt(self, t):
        self.fill("while ")
        self.dispatch(t.condition)
        self.enter()
        self.dispatch(t.body)
        self.leave()
        if t.elsebody:
            self.fill("else")
            self.enter()
            self.dispatch(t.elsebody)
            self.leave()

    def _WithStmt(self, t):
        self.fill("with ")
        first = True
        for item, alias in t.items:
            if first:
                first = False
            else:
                self.write(", ")
            self.dispatch(item)
            if alias:
                self.write(" as ")
                self.dispatch(alias)
        self.enter()
        self.dispatch(t.body)
        self.leave()

    # expr
    def _SimpleExpr(self, t):
        if isinstance(t.value, DistNode):
            self.dispatch(t.value)
        elif t.value:
            self.write(repr(t.value))

    def _ConstantExpr(self, t):
        if isinstance(t.value, int) or isinstance(t.value, float):
            # Substitute overflowing decimal literal for AST infinities.
            self.write(repr(t.value).replace("inf", INFSTR))
        else:
            self.write(repr(t.value))

    def _SelfExpr(self, t):
        self.write("self")

    def _TrueExpr(self, t):
        self.write("True")

    def _FalseExpr(self, t):
        self.write("False")

    def _NoneExpr(self, t):
        self.write("None")

    def _NamedVar(self, t):
        s = t.scope
        if isinstance(s, Process):
            self.write("self.")
        self.write(t.name)

    def _ListExpr(self, t):
        self.write("[")
        interleave(lambda: self.write(", "), self.dispatch, t.subexprs)
        self.write("]")

    def _TupleExpr(self, t):
        self.write("(")
        if len(t.subexprs) == 1:
            (elt,) = t.subexprs
            self.dispatch(elt)
            self.write(",")
        else:
            interleave(lambda: self.write(", "), self.dispatch, t.subexprs)
        self.write(")")

    def _SetExpr(self, t):
        assert(t.subexprs) # should be at least one element
        self.write("{")
        interleave(lambda: self.write(", "), self.dispatch, t.subexprs)
        self.write("}")

    def _DictExpr(self, t):
        assert len(t.keys) == len(t.values)
        self.write("{")
        def write_pair(pair):
            (k, v) = pair
            self.dispatch(k)
            self.write(": ")
            self.dispatch(v)
        interleave(lambda: self.write(", "), write_pair, zip(t.keys, t.values))
        self.write("}")

    def _IfExpr(self, t):
        self.write("(")
        self.dispatch(t.body)
        self.write(" if ")
        self.dispatch(t.condition)
        self.write(" else ")
        self.dispatch(t.orbody)
        self.write(")")

    def _GeneratorExpr(self, t):
        self.write("(")
        self.dispatch(t.elem)
        self.write(" | ")
        interleave(lambda: self.write(", "), self.dispatch, t.conditions)
        self.write(")")

    def _ListCompExpr(self, t):
        self.write("[")
        self.dispatch(t.elem)
        self.write(" | ")
        interleave(lambda: self.write(", "), self.dispatch, t.conditions)
        self.write("]")

    def _SetCompExpr(self, t):
        self.write("{")
        self.dispatch(t.elem)
        self.write(" | ")
        interleave(lambda: self.write(", "), self.dispatch, t.conditions)
        self.write("}")

    def _TupleCompExpr(self, t):
        self.write("()")
        self.dispatch(t.elem)
        self.write(" | ")
        interleave(lambda: self.write(", "), self.dispatch, t.conditions)
        self.write(")")

    def _DictCompExpr(self, t):
        self.write("{")
        self.dispatch(t.elem)
        self.write(" | ")
        interleave(lambda: self.write(", "), self.dispatch, t.conditions)
        self.write("}")

    def _KeyValue(self, t):
        self.dispatch(t.key)
        self.write(": ")
        self.dispatch(t.value)

    def _MaxExpr(self, t):
        self.write("max(")
        self._callargs(t)
        self.write(")")

    def _MinExpr(self, t):
        self.write("min(")
        self._callargs(t)
        self.write(")")

    def _SizeExpr(self, t):
        self.write("size(")
        self._callargs(t)
        self.write(")")

    def _SumExpr(self, t):
        self.write("sum(")
        self._callargs(t)
        self.write(")")

    def _DomainSpec(self, t):
        if not isinstance(t.domain, HistoryExpr):
            self.dispatch(t.pattern)
            self.write("∈")
        self.dispatch(t.domain)

    def _QuantifiedExpr(self, t):
        prefix = QUANTOP_SYMBOLS[t.operator]
        self.write("{")
        for dom in t.domains:
            self.write(prefix)
            self.dispatch(dom)
        self.write(" | ")
        self.dispatch(t.predicate)
        self.write("}")

    def _LogicalExpr(self, t):
        opsym = BOOLOP_SYMBOLS[t.operator]
        if t.operator is NotOp:
            self.write(opsym)
            self.dispatch(t.left)
        else:
            s = " " + opsym + " "
            interleave(lambda: self.write(s), self.dispatch, t.subexprs)

    def _UnaryExpr(self, t):
        self.write("(")
        self.write(UNARYOP_SYMBOLS[t.operator])
        self.write(" ")
        self.dispatch(t.right)
        self.write(")")

    def _BinaryExpr(self, t):
        self.write("(")
        self.dispatch(t.left)
        self.write(" " + BINOP_SYMBOLS[t.operator] + " ")
        self.dispatch(t.right)
        self.write(")")

    def _ComparisonExpr(self, t):
        self.write("(")
        # XXX: Hack! if RHS is HistoryExpr, ignore LHS
        if not isinstance(t.right, HistoryExpr):
            self.dispatch(t.left)
            self.write(" " + CMPOP_SYMBOLS[t.comparator] + " ")
        self.dispatch(t.right)
        self.write(")")

    def _PatternExpr(self, t):
        self.dispatch(t.pattern)

    _LiteralPatternExpr = _PatternExpr

    def _ReceivedExpr(self, t):
        self.write("received(")
        self.dispatch(t.event)
        self.write(")")

    def _SentExpr(self, t):
        self.write("sent(")
        self.dispatch(t.event)
        self.write(")")

    def _CallExpr(self, t):
        if isinstance(t.func, DistNode):
            self.dispatch(t.func)
        else:
            self.write(t.func)
        self.write("(")
        self._callargs(t)
        self.write(")")

    _BuiltinCallExpr = _CallExpr
    _ApiCallExpr = _CallExpr

    def _AttributeExpr(self,t):
        self.dispatch(t.value)
        # Special case: 3.__abs__() is a syntax error, so if t.value
        # is an integer literal then we need to either parenthesize
        # it or add an extra space to get 3 .__abs__().
        if isinstance(t.value, ConstantExpr) and \
           isinstance(t.value.value, int):
            self.write(" ")
        self.write(".")
        self.write(t.attr)

    def _SubscriptExpr(self, t):
        self.dispatch(t.value)
        self.write("[")
        self.dispatch(t.index)
        self.write("]")

    def _SliceExpr(self, t):
        self.write("[")
        if t.lower:
            self.dispatch(t.lower)
        self.write(":")
        if t.upper:
            self.dispatch(t.upper)
        if t.step:
            self.write(":")
            self.dispatch(t.step)
        self.write("]")

    def _StarredExpr(self, t):
        self.write("*")
        self.dispatch(t.value)

    # slice
    def _EllipsisExpr(self, t):
        self.write("...")

    # argument
    def _arg(self, t):
        self.write(t.arg)
        if t.annotation:
            self.write(": ")
            self.dispatch(t.annotation)

    # pattern
    def _ConstantPattern(self, t):
        self.dispatch(t.value)

    def _FreePattern(self, t):
        if t.value:
            self.dispatch(t.value)
        else:
            self.write('_')

    def _BoundPattern(self, t):
        self.write("=")
        self.dispatch(t.value)

    _TuplePattern = _TupleExpr
    _ListPattern = _ListExpr

    # others
    def _Arguments(self, t):
        first = True
        # normal arguments
        defaults = [None] * (len(t.args) - len(t.defaults)) + t.defaults
        for a, d in zip(t.args, defaults):
            if first: first = False
            else: self.write(", ")
            self.dispatch(a)
            if d:
                self.write("=")
                self.dispatch(d)

        # varargs, or bare '*' if no varargs but keyword-only arguments present
        if t.vararg or t.kwonlyargs:
            if first:first = False
            else: self.write(", ")
            self.write("*")
            if t.vararg:
                self.dispatch(t.vararg)

        # keyword-only arguments
        if t.kwonlyargs:
            for a, d in zip(t.kwonlyargs, t.kw_defaults):
                if first:first = False
                else: self.write(", ")
                self.dispatch(a),
                if d:
                    self.write("=")
                    self.dispatch(d)

        # kwargs
        if t.kwarg:
            if first:first = False
            else: self.write(", ")
            self.write("**")
            self.dispatch(t.kwarg.arg)

    def _Event(self, t):
        self.dispatch(t.pattern)
        if t.sources:
            self.write(" from=(")
            self.dispatch(t.sources)
            self.write(")")
        if t.destinations:
            self.write(" to=(")
            self.dispatch(t.destinations)
            self.write(")")
        if t.timestamps:
            self.write(" clk=(")
            self.dispatch(t.timestamps)
            self.write(")")

    def _LambdaExpr(self, t):
        self.write("(")
        self.write("lambda ")
        self.dispatch(t.args)
        self.write(": ")
        self.dispatch(t.body)
        self.write(")")

    def _Alias(self, t):
        if t.asname:
            self.write(t.name + " as ")
            self.dispatch(t.asname)
        else:
            self.dispatch(t.name)

    def _callargs(self, t):
        comma = False
        for e in t.args:
            if comma: self.write(", ")
            else: comma = True
            self.dispatch(e)
        for key, value in t.keywords:
            if comma: self.write(", ")
            else: comma = True
            self.write(key)
            self.write("=")
            self.dispatch(value)
        if t.starargs:
            if comma: self.write(", ")
            else: comma = True
            self.write("*")
            self.dispatch(t.starargs)
        if t.kwargs:
            if comma: self.write(", ")
            else: comma = True
            self.write("**")
            self.dispatch(t.kwargs)

    def _withitem(self, t):
        self.dispatch(t.context_expr)
        if t.optional_vars:
            self.write(" as ")
            self.dispatch(t.optional_vars)

if __name__ == '__main__':
    from da.compiler.ui import daast_from_file
    from da.compiler.utils import to_pseudo
    ast = daast_from_file("../test/await.da")
    print(to_pseudo(ast))



# -*- generated by 1.0.6 -*-
import da
PatternExpr_189 = da.pat.TuplePattern([da.pat.ConstantPattern('request'), da.pat.FreePattern('c2'), da.pat.FreePattern('p')])
PatternExpr_214 = da.pat.TuplePattern([da.pat.ConstantPattern('release'), da.pat.BoundPattern('_BoundPattern217_'), da.pat.BoundPattern('_BoundPattern218_')])
PatternExpr_247 = da.pat.TuplePattern([da.pat.ConstantPattern('ack'), da.pat.BoundPattern('_BoundPattern250_'), da.pat.BoundPattern('_BoundPattern251_')])
PatternExpr_279 = da.pat.TuplePattern([da.pat.ConstantPattern('request'), da.pat.FreePattern('c'), da.pat.FreePattern('p')])
PatternExpr_330 = da.pat.TuplePattern([da.pat.ConstantPattern('done'), da.pat.BoundPattern('_BoundPattern333_')])
PatternExpr_336 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('done'), da.pat.BoundPattern('_BoundPattern346_')])])
PatternExpr_221 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('release'), da.pat.BoundPattern('_BoundPattern231_'), da.pat.BoundPattern('_BoundPattern232_')])])
PatternExpr_254 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('ack'), da.pat.BoundPattern('_BoundPattern264_'), da.pat.BoundPattern('_BoundPattern265_')])])
_config_object = {'channel': 'fifo', 'clock': 'Lamport'}
import sys

class P(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._PReceivedEvent_0 = []
        self._PReceivedEvent_1 = []
        self._PReceivedEvent_2 = []
        self._PReceivedEvent_4 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_0', PatternExpr_189, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_1', PatternExpr_214, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_2', PatternExpr_247, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_3', PatternExpr_279, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._P_handler_278]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_4', PatternExpr_330, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, s, nrequests, **rest_409):
        super().setup(s=s, nrequests=nrequests, **rest_409)
        self._state.s = s
        self._state.nrequests = nrequests
        pass

    def run(self):

        def task():
            self.output('in cs')
            self.output('releasing cs')
        for i in range(self._state.nrequests):
            self.mutex(task)
        self.send(('done', self._id), to=self._state.s)
        super()._label('_st_label_321', block=False)
        p = None

        def UniversalOpExpr_322():
            nonlocal p
            for p in self._state.s:
                if (not PatternExpr_336.match_iter(self._PReceivedEvent_4, _BoundPattern346_=p, SELF_ID=self._id)):
                    return False
            return True
        _st_label_321 = 0
        while (_st_label_321 == 0):
            _st_label_321 += 1
            if UniversalOpExpr_322():
                _st_label_321 += 1
            else:
                super()._label('_st_label_321', block=True)
                _st_label_321 -= 1
        self.output('terminating')

    def mutex(self, task):
        super()._label('request', block=False)
        c = self.logical_clock()
        self.send(('request', c, self._id), to=self._state.s)
        super()._label('_st_label_185', block=False)
        p = c2 = None

        def UniversalOpExpr_187():
            nonlocal p, c2
            for (_, _, (_ConstantPattern206_, c2, p)) in self._PReceivedEvent_0:
                if (_ConstantPattern206_ == 'request'):
                    if (not (PatternExpr_221.match_iter(self._PReceivedEvent_1, _BoundPattern231_=c2, _BoundPattern232_=p, SELF_ID=self._id) or ((c, self._id) < (c2, p)))):
                        return False
            return True
        p = None

        def UniversalOpExpr_240():
            nonlocal p
            for p in self._state.s:
                if (not PatternExpr_254.match_iter(self._PReceivedEvent_2, _BoundPattern264_=c, _BoundPattern265_=p, SELF_ID=self._id)):
                    return False
            return True
        _st_label_185 = 0
        while (_st_label_185 == 0):
            _st_label_185 += 1
            if (UniversalOpExpr_187() and UniversalOpExpr_240()):
                _st_label_185 += 1
            else:
                super()._label('_st_label_185', block=True)
                _st_label_185 -= 1
        super()._label('critical_section', block=False)
        task()
        super()._label('release', block=False)
        self.send(('release', c, self._id), to=self._state.s)

    def _P_handler_278(self, c, p):
        self.send(('ack', c, self._id), to=p)
    _P_handler_278._labels = None
    _P_handler_278._notlabels = None

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])

    def run(self):
        nprocs = (int(sys.argv[1]) if (len(sys.argv) > 1) else 10)
        nrequests = (int(sys.argv[2]) if (len(sys.argv) > 2) else 1)
        ps = self.new(P, num=nprocs)
        for p in ps:
            self._setup(p, ((ps - {p}), nrequests))
        self._start(ps)



# Copyright (c) 2010-2016 Bo Lin
# Copyright (c) 2010-2016 Yanhong Annie Liu
# Copyright (c) 2010-2016 Stony Brook University
# Copyright (c) 2010-2016 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import sys
import time
import stat
import logging
import collections.abc
import os.path

from sys import stderr
from . import common, sim, endpoint
from .common import api, deprecated, get_runtime_option, ProcessId

DISTPY_SUFFIXES = [".da", ""]
PYTHON_SUFFIX = ".py"
NODECLS = "Node_"
BASE_MODULE_NAME = 'da.lib.base'
DEFAULT_MASTER_PORT = 15000
ASYNC_TIMEOUT = 5
PORT_RANGE = 10

CONSOLE_LOG_FORMAT = \
    '[%(relativeCreated)d] %(name)s%(daPid)s:%(levelname)s: %(message)s'
FILE_LOG_FORMAT = \
    '[%(asctime)s] %(name)s%(daPid)s:%(levelname)s: %(message)s'
HEADER_REGEXP = "# -\*- generated by ([0-9a-z.-]*) -\*-"

log = logging.getLogger(__name__)

def find_file_on_paths(filename, paths):
    """Looks for a given 'filename' under a list of directories, in order.

    If found, returns a pair (path, mode), where 'path' is the full path to
    the file, and 'mode' is the result of calling 'os.stat' on the file.
    Otherwise, returns (None, None).

    """
    for path in paths:
        fullpath = os.path.join(path, filename)
        try:
            filemode = os.stat(fullpath)
            return fullpath, filemode
        except OSError:
            pass
    return None, None

def strip_suffix(filename):
    """Returns a filename minus it's extension."""

    dotidx = filename.rfind(".")
    return filename[:dotidx] if dotidx != -1 else filename

@api
def init(**configs):
    """Initializes the DistAlgo runtime.

    """
    common.initialize_runtime_options(configs)

@api
def import_da(name, from_dir=None, compiler_args=[]):
    """Imports DistAlgo module 'module', returns the module object.

    This function mimics the Python builtin __import__() function for DistAlgo
    modules. 'name' is the name of the module to be imported, in
    "dotted module name" format. The module must be implemented in one regular
    DistAlgo source file with a '.da' filename suffix; package modules are not
    supported.

    This function returns the imported module object upon successful import;
    otherwise, 'ImportError' is raised.

    Optional argument 'compiler_args' is a list of command line arguments to
    pass to the compiler, if a compiler invocation is needed. Optional
    argument 'from_dir' should be a valid module search path that overrides
    'sys.path'.

    """
    import re
    import importlib
    import da.compiler as compiler

    force_recompile = get_runtime_option('recompile', default=False)
    paths = sys.path if from_dir is None else [from_dir]
    pathname = name.replace(".", os.sep)
    for suffix in DISTPY_SUFFIXES:
        fullpath, mode = find_file_on_paths(pathname + suffix, paths)
        if fullpath is not None:
            break
    if fullpath is None:
        raise ImportError("Module %s not found." % name)
    pyname = strip_suffix(fullpath) + PYTHON_SUFFIX
    try:
        pymode = os.stat(pyname)
        with open(pyname, "r") as fd:
            header = fd.readline(80)
        res = re.match(HEADER_REGEXP, header)
        if res is None or (res.group(1) != common.__version__):
            force_recompile = True
    except OSError:
        pymode = None

    if (force_recompile or pymode is None or
            pymode[stat.ST_MTIME] < mode[stat.ST_MTIME]):
        oldargv = sys.argv
        try:
            argv = oldargv[0:0] + compiler_args + [fullpath]
            res = compiler.ui.main(argv)
        except Exception as err:
            raise RuntimeError("Compiler failure!", err)
        finally:
            sys.argv = oldargv

        if res != 0:
            raise ImportError("Unable to compile %s, errno: %d" %
                              (fullpath, res))

    moduleobj = importlib.import_module(name)
    common.setup_logging_for_module(name, CONSOLE_LOG_FORMAT, FILE_LOG_FORMAT)
    return moduleobj

def _load_cookie():
    authkey = get_runtime_option('cookie')
    if authkey is None:
        try:
            fname = os.path.expanduser("~/.da.cookie")
            with open(fname, "r") as fd:
                authkey = fd.read(80).encode()
        except OSError:
            pass
    return authkey

def _parse_address(straddr):
    assert isinstance(straddr, str)
    components = straddr.split(':')
    if len(components) > 2:
        raise ValueError("Invalid address: {}".format(straddr))
    elif len(components) == 0:
        return "", None
    elif len(components) == 1:
        return components[0], None
    else:
        try:
            return components[0], int(components[1])
        except ValueError as e:
            raise ValueError("Invalid port number: {}".format(components[1]))

def _bootstrap_node(cls, nodename, trman):
    router = None
    is_master = get_runtime_option('master')
    hostname = get_runtime_option('hostname')
    port = get_runtime_option('port')
    if port is None:
        port = get_runtime_option('default_master_port')
        strict = False
        if is_master:
            try:
                trman.initialize(hostname=hostname, port=port, strict=True)
            except endpoint.TransportException as e:
                log.debug("Binding attempt to port %d failed: %r", port, e)
                trman.close()
    else:
        strict = True
    if not trman.initialized:
        trman.initialize(hostname=hostname, port=port,
                         strict=strict, linear=is_master)
    nid = ProcessId._create(pcls=cls,
                            transports=trman.transport_addresses,
                            name=nodename)
    common._set_node(nid)
    if not is_master:
        rhost, rport = _parse_address(get_runtime_option('peer'))
        if len(rhost) == 0:
            rhost = hostname
        if rport is None:
            rport = get_runtime_option('default_master_port')
        trman.start()
        router = sim.Router(trman)
        try:
            router.bootstrap_node(rhost, rport, timeout=ASYNC_TIMEOUT)
        except sim.BootstrapException as e:
            log.info("Bootstrapping attempt failed due to %r, "
                     " continuing as a master node (use '--master' to "
                     "disable bootstrapping at startup).", e)
        router.start()
    return router

def _load_main_module():
    target = get_runtime_option('file')
    compiler_args = get_runtime_option('compiler_flags').split()
    if target is not None:
        source_dir = os.path.dirname(target)
        basename = strip_suffix(os.path.basename(target))
        if not os.access(target, os.R_OK):
            die("Can not access source file %s" % target)
        sys.path.insert(0, source_dir)
        module = import_da(basename,
                           from_dir=source_dir,
                           compiler_args=compiler_args)
        sys.argv = [target] + get_runtime_option('args')
    elif get_runtime_option('module') is not None:
        module_args = get_runtime_option('module')
        module_name = module_args[0]
        module = import_da(module_name, compiler_args=compiler_args)
        sys.argv = ['__main__'] + module_args[1:]
    else:
        module = import_da(BASE_MODULE_NAME)
    return module

def _check_nodename():
    nodename = get_runtime_option('nodename')
    if not common.check_name(nodename):
        die("'--nodename' must not contain any of the characters in {}".
            format(common.ILLEGAL_NAME_CHARS))
    return nodename

def entrypoint():
    """Entry point for running DistAlgo as the main module.

    """
    import socket
    import multiprocessing

    if sys.version_info < (3, 4):
        die("DistAlgo requires Python version 3.4 or newer.\n")

    startmeth = get_runtime_option('start_method')
    if startmeth != multiprocessing.get_start_method(allow_none=True):
        multiprocessing.set_start_method(startmeth)
    try:
        module = _load_main_module()
    except ImportError as e:
        die("ImportError: " + str(e))
    is_idle = get_runtime_option('idle')
    if not hasattr(module, 'Node_'):
        if is_idle:
            # Just use the generic node:
            module.Node_ = sim.NodeProcess
        else:
            die("Main process not defined!")
    elif not (type(module.Node_) is type and
              issubclass(module.Node_, sim.DistProcess)):
        die("Main process is not a DistProcess: {}".format(module.Node_))

    common.set_runtime_option('this_module_name', module.__name__)
    common.set_runtime_option('main_module_name', module.__name__)
    if get_runtime_option('inc_module_name') is None:
        common.set_runtime_option('inc_module_name', module.__name__ + "_inc")
    hostname = socket.gethostbyname(common.get_runtime_option('hostname'))
    common.set_runtime_option('hostname', hostname)
    common.sysinit()
    common.setup_logging_for_module("da")

    # Start main program
    nodename = _check_nodename()
    niters = get_runtime_option('iterations')
    if len(nodename) == 0:
        # Safety precaution: disallow distributed messaging when no node name
        # specified:
        cookie = multiprocessing.current_process().authkey
    else:
        cookie = _load_cookie()
    nodeimpl = None
    router = None
    trman = None
    try:
        trman = endpoint.TransportManager(cookie)
        if len(nodename) > 0:
            router = _bootstrap_node(module.Node_, nodename, trman)
            nid = common.pid_of_node()
        else:
            trman.initialize()
            nid = ProcessId._create(pcls=module.Node_,
                                    transports=trman.transport_addresses,
                                    name=nodename)
            common._set_node(nid)

        log.info("%s initialized at %s:(%s).", nid,
                 get_runtime_option('hostname'), trman.transport_addresses_str)
        log.info("Starting program %s...", module)
        for i in range(0, niters):
            log.info("Running iteration %d ...", (i+1))

            nodeimpl = sim.OSThreadContainer(process_class=module.Node_,
                                             transport_manager=trman,
                                             process_id=nid,
                                             parent_id=nid,
                                             process_name=nodename,
                                             router=router)
            nodeimpl.start()
            log.info("Waiting for remaining child processes to terminate..."
                     "(Press \"Ctrl-C\" to force kill)")
            nodeimpl.join()
            nodeimpl = None
            log.info("Main process terminated.")
        return 0

    except (endpoint.TransportException, sim.RoutingException) as e:
        log.error("Transport initialization failed due to: %r", e)
        stderr.write("System failed to start. \n")
        return 5
    except KeyboardInterrupt as e:
        log.warning("Received keyboard interrupt. ")
        if nodeimpl is not None:
            stderr.write("Terminating node...")
            nodeimpl.end()
            t = 0
            while nodeimpl.is_alive() and t < ASYNC_TIMEOUT:
                stderr.write(".")
                t += 1
                nodeimpl.join(timeout=1)
        if nodeimpl is not None and nodeimpl.is_alive():
            stderr.write("\nNode did not terminate gracefully, "
                         "some zombie child processes may be present.\n")
            return 2
        else:
            stderr.write("\nNode terminated. Goodbye!\n")
            return 1
    except Exception as e:
        log.error("Caught unexpected global exception: %r", e, exc_info=1)
        return 4

def die(mesg = None):
    if mesg != None:
        stderr.write(mesg + "\n")
    sys.exit(1)



# Copyright (c) 2010-2016 Bo Lin
# Copyright (c) 2010-2016 Yanhong Annie Liu
# Copyright (c) 2010-2016 Stony Brook University
# Copyright (c) 2010-2016 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import enum
import time
import random
import socket
import logging
import threading
import selectors
import multiprocessing
from collections import namedtuple

from . import common
from .common import get_runtime_option, LRU, WaitableQueue

logger = logging.getLogger(__name__)

HEADER_SIZE = 8
BYTEORDER = 'big'
MAX_RETRY = 3

class ChannelCaps:
    """An enum of channel capabilities."""
    FIFO         = 1
    RELIABLE     = 2
    INTERHOST    = 4
    BROADCAST    = 8
    RELIABLEFIFO = FIFO | RELIABLE

class TransportException(Exception): pass
class VersionMismatchException(TransportException): pass
class AuthenticationException(TransportException): pass
class BindingException(TransportException): pass
class NoAvailablePortsException(BindingException): pass
class NoTargetTransportException(TransportException): pass
class InvalidTransportStateException(TransportException): pass
class PacketSizeExceededException(TransportException): pass

TransportTypes = []

class TransportManager:
    """Manages all DistAlgo transports within a process.

    """
    log = None

    def __init__(self, cookie=None):
        self.queue = None
        self.transports = []
        self.initialized = False
        self.started = False
        self.authkey = cookie
        if self.__class__.log is None:
            self.__class__.log = logger.getChild(self.__class__.__qualname__)

    def __getstate__(self):
        return (self.initialized, self.started, self.authkey)

    def __setstate__(self, state):
        self.initialized, self.started, self.authkey = state
        if self.__class__.log is None:
            self.__class__.log = logger.getChild(self.__class__.__qualname__)

    @property
    def transport_addresses(self):
        return tuple(t.address for t in self.transports)

    @property
    def transport_addresses_str(self):
        return ", ".join(["{}={}".format(tr.__class__.__name__, tr.address)
                        for tr in self.transports])

    def initialize(self, pipe=None, **params):
        """Initialize all transports.

        """
        self.log.debug("Initializing with key %r...", self.authkey)
        total = len(TransportTypes)
        self.transports = tuple(cls(self.authkey) for cls in TransportTypes)
        cnt = 0
        for transport in self.transports:
            try:
                if pipe is not None:
                    assert pipe.recv() is transport.__class__
                transport.initialize(pipe=pipe, **params)
                cnt += 1
            except Exception as err:
                self.log.debug("Failed to initialize transport %s: %r",
                               transport, err, exc_info=1)
        if pipe:
            pipe.send('done')
        if cnt != total:
            raise TransportException(
                "Initialization failed for {}/{} transports.".format(
                    (total - cnt), total))
        else:
            self.initialized = True

    def start(self):
        """Start all transports.

        """
        self.log.debug("Starting...")
        self.queue = WaitableQueue()
        total = len(TransportTypes)
        cnt = 0
        for transport in self.transports:
            try:
                transport.start(self.queue)
                cnt += 1
            except Exception as err:
                self.log.error("Failed to start transport %s: %r", transport, err)
        if cnt != total:
            raise TransportException(
                "Start failed for {}/{} transports.".format(
                    (total - cnt), total))
        else:
            self.started = True

    def close(self):
        """Shut down all transports.

        """
        self.log.debug("Stopping...")
        total = len(TransportTypes)
        cnt = 0
        for transport in self.transports:
            try:
                transport.close()
                cnt += 1
            except Exception as err:
                self.log.warning("Exception when stopping transport %s: %r",
                                 transport, err)
        self.started = False
        self.initialized = False
        self.log.debug("%d/%d transports stopped.", cnt, total)

    def serialize(self, pipe, pid):
        """Sends all transports to child process.
        """
        for transport in self.transports:
            pipe.send(transport.__class__)
            transport.serialize(pipe, pid)

    def get_transport(self, flags):
        """Returns the first transport instance satisfying `flags`, or None if
        no transport satisfies `flags`.
        """
        flags &= ~(ChannelCaps.BROADCAST)
        for tr in self.transports:
            if (flags & tr.capabilities) == 0:
                return tr
        return None


class Transport:
    """Represents a type of communication channel for sending of data.

    This is the base class for all types of communication channels in DistAlgo.

    """
    slot_index = 0
    def __init__(self, authkey):
        super().__init__()
        self._log = logger.getChild(self.__class__.__qualname__)
        self.queue = None
        self.hostname = None
        self.authkey = authkey

    def initialize(self, hostname=None, **params):
        if hostname is None:
            hostname = get_runtime_option('hostname')
        self.hostname = hostname

    def start(self, queue):
        """Starts the transport.

        """
        self.queue = queue

    def close(self):
        """Stops the transport and clean up resources.

        """
        pass

    def send(self, data, dest, **params):
        """Send `data` to `dest`.

        `data` should be a `bytes` or `bytearray` object. `dest` should be a
        DistAlgo process id.

        """
        pass

    def setname(self, name):
        self._name = name

    @classmethod
    def address_from_id(cls, target):
        """Returns the transport address of `target`.

        Given process id `target`, return the address of it's corresponding
        transport, or None if `target` does not have a corresponding transport.

        """
        address = target.transports[cls.slot_index]
        if address is None:
            return None
        else:
            return (target.hostname, address)

    @property
    def address(self):
        return None

    def __str__(self):
        return self.__class__.__name__

def transport(cls):
    """Decorator to register `cls` as a transport.

    """
    cls.slot_index = len(TransportTypes)
    TransportTypes.append(cls)
    return cls

class SocketTransport(Transport):
    """Base class for socket-based transports.

    """
    capabilities = ~(ChannelCaps.INTERHOST)
    def __init__(self, authkey):
        super().__init__(authkey)
        self.port = None
        self.conn = None
        self.worker = None
        self.buffer_size = 0

    def initialize(self, port=None, strict=False, linear=False,
                   retries=MAX_RETRY, **rest):
        super().initialize(**rest)
        self.buffer_size = get_runtime_option('message_buffer_size')
        assert self.conn is not None
        try:
            _, bound_port = self.conn.getsockname()
            if bound_port != 0:
                # We've already inherited the socket from the parent
                self.port = bound_port
                return
        except OSError as e:
            # This is what we get on Windows if we call `getsockname()` on an
            # unbound socket...
            pass
        self.port = port
        if self.port is None:
            if not strict:
                self.port = random.randint(MIN_TCP_PORT, MAX_TCP_PORT)
            else:
                raise NoAvailablePortsException("Port number not specified!")
        address = None
        retry = 1
        while True:
            address = (self.hostname, self.port)
            try:
                self.conn.bind(address)
                break
            except socket.error as e:
                address = None
                if not strict and retry < retries:
                    if linear:
                        self.port += 1
                    else:
                        self.port = random.randint(MIN_TCP_PORT, MAX_TCP_PORT)
                    retry += 1
                else:
                    raise BindingException(
                        "Failed to bind to an available port.") from e
        self._log.debug("Transport initialized at address %s", address)

    def serialize(self, pipe, pid):
        from multiprocessing.reduction import send_handle
        send_handle(pipe, self.conn.fileno(), pid)

    def close(self):
        if self.conn is None:
            self._log.debug("Already stopped.")
        else:
            try:
                self.conn.close()
            except OSError:
                pass
            finally:
                self.conn = None
        # No need to care about the worker thread since it's a daemon:
        self.worker = None
        self._log.debug("Transport stopped.")

    @property
    def address(self):
        return self.port

    def __str__(self):
        fmt = "<{0.__class__.__qualname__}({0.hostname}:{0.port})>"
        return fmt.format(self)


# UDP Implementation:

MIN_UDP_PORT = 10000
MAX_UDP_PORT = 40000
DIGEST_LENGTH = 16
DIGEST_HOLDER = b'0' * DIGEST_LENGTH

@transport
class UdpTransport(SocketTransport):
    """A channel that supports sending and receiving messages via UDP.

    """
    capabilities = ~(ChannelCaps.INTERHOST)
    data_offset = 4 + DIGEST_LENGTH

    def __init__(self, authkey):
        super().__init__(authkey)

    def initialize(self, pipe=None, **params):
        try:
            if pipe is None:
                self.conn = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                self.conn.set_inheritable(True)
            else:
                from multiprocessing.reduction import recv_handle
                self.conn = socket.fromfd(recv_handle(pipe),
                                          socket.AF_INET, socket.SOCK_DGRAM)
            super().initialize(**params)
        except Exception as e:
            if self.conn is not None:
                self.conn.close()
            self.conn = None
            raise e

    def start(self, queue):
        if self.conn is None:
            raise InvalidTransportStateException(
                "Transport has not been initialized!")
        self.queue = queue
        if self.worker is None or not self.worker.is_alive():
            self.worker = threading.Thread(target=self.recvmesgs, daemon=True)
            self.worker.start()
        self._log.debug("Transport started.")

    def close(self):
        super().close()

    def _packet_from(self, chunk):
        if self.authkey is not None:
            import hmac
            digest = hmac.new(self.authkey, chunk, 'md5').digest()
            return (common.VERSION_BYTES, digest, chunk)
        else:
            return (common.VERSION_BYTES, DIGEST_HOLDER, chunk)

    def _verify_packet(self, chunk, addr):
        if chunk[:4] != common.VERSION_BYTES:
            raise VersionMismatchException("wrong version: {}".format(chunk[:4]))
        if self.authkey is not None:
            with memoryview(chunk)[self.data_offset:] as data:
                import hmac
                digest = hmac.new(self.authkey, data, 'md5').digest()
                if digest != chunk[4:self.data_offset]:
                    raise AuthenticationException(
                        "wrong digest from {}: {}"
                        .format(addr, chunk[4:self.data_offset]))
        else:
            if chunk[4:self.data_offset] != DIGEST_HOLDER:
                raise AuthenticationException('{} requires a cookie.'
                                              .format(addr))

    def _sendmsg_nix(self, packet, target):
        packet_size = sum(len(e) for e in packet)
        return self.conn.sendmsg(packet, [], 0, target) == packet_size

    def _sendmsg_nt(self, packet, target):
        from itertools import chain
        buf = bytes(chain(*packet))
        return self.conn.sendto(buf, target) == len(buf)

    if sys.platform == 'win32':
        _sendmsg = _sendmsg_nt
    else:
        _sendmsg = _sendmsg_nix

    def send(self, chunk, dest, wait=0.01, retries=MAX_RETRY, **rest):
        if self.conn is None:
            raise InvalidTransportStateException(
                "Invalid transport state for sending.")

        if len(chunk) > self.buffer_size:
            self._log.warning("Data size exceeded maximum buffer size!"
                              " Outgoing packet dropped.")
            self._log.debug("Dropped packet: %s", chunk)
            raise PacketSizeExceededException()
        else:
            target = self.address_from_id(dest)
            if target is None:
                raise NoTargetTransportException()
            packet = self._packet_from(chunk)
            cnt = 0
            while True:
                try:
                    if self._sendmsg(packet, target):
                        return
                    else:
                        raise TransportException("Unable to send full chunk.")
                except PermissionError as e:
                    # The 'conntrack' module of iptables will cause UDP `sendto`
                    # to return `EPERM` if it's sending too fast:
                    self._log.debug("Packet to %s dropped by iptables, "
                                    "reduce send rate.", target)
                    cnt += 1
                    if cnt >= retries:
                        raise TransportException("Packet blocked by OS.") from e
                    else:
                        time.sleep(wait)

    def _recvmsg_nt(self):
        chunk, remote = self.conn.recvfrom(self.buffer_size)
        return chunk, None, 0, remote

    def _recvmsg_nix(self):
        return self.conn.recvmsg(self.buffer_size)

    if sys.platform == 'win32':
        _recvmsg = _recvmsg_nt
        socket.MSG_ERRQUEUE = 0
    elif sys.platform == 'darwin':
        _recvmsg = _recvmsg_nix
        socket.MSG_ERRQUEUE = 0
    else:
        _recvmsg = _recvmsg_nix

    def recvmesgs(self):
        if self.conn is None:
            raise InvalidTransportStateException(
                "Invalid transport state for receiving.")

        try:
            while True:
                chunk, _, flags, remote= self._recvmsg()
                if not chunk:
                    # XXX: zero length packet == closed socket??
                    self._log.debug("Transport closed, terminating receive loop.")
                    break
                elif flags & socket.MSG_TRUNC:
                    self._log.debug("Dropped truncated packet. ")
                elif flags & socket.MSG_ERRQUEUE:
                    self._log.debug("No data received. ")
                else:
                    try:
                        self._verify_packet(chunk, remote)
                        self.queue.append((self.__class__, chunk, remote))
                    except TransportException as e:
                        self._log.warning("Packet from %s dropped due to: %r",
                                          remote, e)
        except (socket.error, AttributeError) as e:
            self._log.debug("Terminating receive loop due to %r", e)

    @property
    def started(self):
        return self.conn is not None


# TCP Implementation:

MAX_TCP_BACKLOG = 10
MAX_TCP_CONN = 200
MIN_TCP_PORT = 10000
MAX_TCP_PORT = 40000

#
# Authentication stuff
#

MESSAGE_LENGTH = 20

KEY_CHALLENGE = b'#KY#'
VER_CHALLENGE = b'#VR#'
WELCOME = b'#WELCOME#'
FAILURE = b'#FAILURE#'

class AuxConnectionData:
    """Auxiliary data associated with each TCP connection.

    """
    def __init__(self, peername, message_size, digest=None, provision=False):
        self.peername = peername
        self.message_size = message_size
        self.digest = digest
        if provision:
            self.provision()

    def provision(self):
        self.buf = bytearray(self.message_size * 2)
        self.view = memoryview(self.buf)
        self.lastptr = 0
        self.freeptr = 0


@transport
class TcpTransport(SocketTransport):
    """A channel that supports sending and receiving messages via TCP.

    """
    capabilities = ~((ChannelCaps.FIFO) |
                     (ChannelCaps.RELIABLE) |
                     (ChannelCaps.INTERHOST))
    data_offset = 0

    def __init__(self, authkey):
        super().__init__(authkey)
        self.cache = None
        self.lock = None
        self.selector = None

    def initialize(self, strict=False, pipe=None, **params):
        try:
            if pipe is None:
                self.conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                if strict and not get_runtime_option('tcp_dont_reuse_addr'):
                    self.conn.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            else:
                from multiprocessing.reduction import recv_handle
                self.conn = socket.fromfd(recv_handle(pipe),
                                          socket.AF_INET, socket.SOCK_STREAM)
            super().initialize(strict=strict, **params)
        except Exception as e:
            if self.conn is not None:
                self.conn.close()
            self.conn = None
            raise e

    def start(self, queue):
        if self.conn is None:
            raise InvalidTransportStateException(
                "Transport has not been initialized!")
        self.queue = queue
        self.lock = threading.Lock()
        self.conn.listen(MAX_TCP_BACKLOG)
        self.conn.settimeout(5)
        self.cache = dict()
        self.selector = selectors.DefaultSelector()
        self.selector.register(self.conn, selectors.EVENT_READ,
                               (self._accept, None))
        if self.worker is None or not self.worker.is_alive():
            self.worker = threading.Thread(target=self.recvmesgs, daemon=True)
            self.worker.start()
        self._log.debug("Transport started.")

    def close(self):
        if self.selector is not None:
            self.selector.close()
        super().close()
        self.cache = None
        self.lock = None
        self.selector = None

    def _deliver_challenge(self, conn, addr):
        import os
        digest = None
        if self.authkey is not None:
            import hmac
            message = os.urandom(MESSAGE_LENGTH)
            self._send_1((KEY_CHALLENGE, common.VERSION_BYTES, message),
                         conn, addr)
            digest = hmac.new(self.authkey, message, 'md5').digest()
        else:
            self._send_1((VER_CHALLENGE, common.VERSION_BYTES), conn, addr)
        return digest

    def _verify_challenge(self, conn, auxdata):
        addr = auxdata.peername
        self.selector.unregister(conn)
        message = conn.recv(256)
        if self.authkey is not None:
            if message != auxdata.digest:
                self._send_1((FAILURE,), conn, addr)
                raise AuthenticationException(
                    'Digest from {0.peername} was wrong.'.format(auxdata))
        else:
            if message == KEY_CHALLENGE:
                raise AuthenticationException(
                    '{0.peername} requires a cookie.'.format(auxdata))
            if message != VER_CHALLENGE:
                raise VersionMismatchException(
                    'Version from {0.peername} is different.'.format(auxdata))
        self._send_1((WELCOME,), conn, addr)
        if auxdata.peername in self.cache:
            self._log.warning("Double connection from %s!", auxdata.peername)
        with self.lock:
            self.cache[auxdata.peername] = conn
        auxdata.provision()
        self.selector.register(conn, selectors.EVENT_READ,
                               (self._receive_1, auxdata))

    def _answer_challenge(self, conn, addr):
        message = conn.recv(256)
        self._log.debug("=========answering %r", message)
        if self.authkey is not None:
            import hmac
            if message[:len(KEY_CHALLENGE)] != KEY_CHALLENGE:
                self._send_1((KEY_CHALLENGE,), conn, addr)
                raise AuthenticationException('{} has no cookie.'.
                                              format(addr))
            if message[len(KEY_CHALLENGE):len(KEY_CHALLENGE)+4] != \
               common.VERSION_BYTES:
                raise VersionMismatchException('Version at {} is different.'.
                                               format(addr))
            message = message[len(KEY_CHALLENGE)+4:]
            digest = hmac.new(self.authkey, message, 'md5').digest()
            self._send_1((digest,), conn, addr)
        else:
            if message[:len(KEY_CHALLENGE)] == KEY_CHALLENGE:
                self._send_1((KEY_CHALLENGE,), conn, addr)
                raise AuthenticationException('{} requires a cookie.'.
                                              format(addr))
            elif message != VER_CHALLENGE + common.VERSION_BYTES:
                self._send_1((FAILURE,), conn, addr)
                raise VersionMismatchException('Version at {} is different.'.
                                               format(addr))
            else:
                self._send_1((VER_CHALLENGE,), conn, addr)
        response = conn.recv(256)
        if response != WELCOME:
            raise AuthenticationException('digest was rejected by {}.'.
                                          format(addr))

    def _accept(self, conn, auxdata):
        conn, addr = self.conn.accept()
        self._log.debug("Accepted connection from %s.", addr)
        digest = self._deliver_challenge(conn, auxdata)
        self.selector.register(conn, selectors.EVENT_READ,
                               (self._verify_challenge,
                                AuxConnectionData(addr,
                                                  self.buffer_size,
                                                  digest)))

    def _connect(self, target):
        self._log.debug("Initiating connection to %s.", target)
        conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        conn.settimeout(5)
        conn.connect(target)
        self._answer_challenge(conn, target)
        self._log.debug("Connection to %s established.", target)
        return conn

    def _cleanup(self, conn, remote):
        self._log.debug("Cleanup connection to %s.", remote)
        try:
            self.selector.unregister(conn)
        except (KeyError, ValueError):
            pass
        if remote in self.cache:
            with self.lock:
                if self.cache[remote] is conn:
                    del self.cache[remote]
                else:
                    self._log.warning("Possible corrupted cache entry for %s",
                                      remote)
        try:
            conn.close()
        except OSError:
            pass

    def send(self, chunk, dest, retries=MAX_RETRY, wait=0.05,
             retry_refused_connections=False, **rest):
        target = self.address_from_id(dest)
        if target is None:
            raise NoTargetTransportException(
                "Process {} does not have TCP transport!".format(dest))

        retry = 1
        saved = conn = None
        with self.lock:
            saved = conn = self.cache.get(target)
        try:
            while True:
                try:
                    if conn is None:
                        conn = self._connect(target)
                    l = len(chunk)
                    header = int(l).to_bytes(HEADER_SIZE, BYTEORDER)
                    self._send_1((header, chunk), conn, target)
                    return
                except ConnectionRefusedError as e:
                    if (not retry_refused_connections) or retry > retries:
                        raise TransportException(
                            'connection refused by {}'.format(target)) from e
                except (socket.error, socket.timeout) as e:
                    self._log.debug("Sending to %s failed on %dth try: %r",
                                    dest, retry, e)
                    if conn is not None:
                        conn.close()
                        conn = None
                    if retry > retries:
                        raise TransportException('max retries reached.') from e
                retry += 1
                time.sleep(wait)
        finally:
            if conn is not None:
                if saved != conn:
                    with self.lock:
                        self.cache[target] = conn
                    self.selector.register(
                        conn, selectors.EVENT_READ,
                        (self._receive_1,
                         AuxConnectionData(target, self.buffer_size,
                                           provision=True)))
            else:
                if target in self.cache:
                    with self.lock:
                        del self.cache[target]

    def _send_1(self, data, conn, target=None):
        msglen = sum(len(chunk) for chunk in data)
        sent = conn.sendmsg(data)
        if sent != msglen:
            self._log.debug("_send_1: only sent %d/%d bytes. ", sent, msglen)
            raise socket.error("Unable to send full chunk.")
        else:
            self._log.debug("Sent %d bytes to %s.", msglen, target)

    def _send_1_nt(self, data, conn, target=None):
        from itertools import chain
        buf = bytes(chain(*data))
        conn.sendall(buf)

    if sys.platform == 'win32':
        _send_1 = _send_1_nt

    def recvmesgs(self):
        if self.conn is None:
            raise InvalidTransportStateException(
                "Invalid transport state for receiving.")
        try:
            while True:
                events = self.selector.select()
                for key, mask in events:
                    callback, aux = key.data
                    try:
                        callback(key.fileobj, aux)
                    except TransportException as e:
                        self._log.warning("Exception when handling %s: %r",
                                          aux.peername, e)
                        if key.fileobj is not self.conn:
                            self._cleanup(key.fileobj, aux.peername)
                    except socket.error as e:
                        if key.fileobj is self.conn:
                            self._log.error("socket.error on listener: %r", e)
                            break
                        else:
                            self._log.debug(
                                "socket.error when receiving from %s: %r",
                                aux.peername, e)
                            self._cleanup(key.fileobj, aux.peername)
        except Exception as e:
            self._log.debug("(recvmesgs): caught exception %r", e, exc_info=1)

    def _receive_1(self, conn, aux):
        buf = aux.buf
        view = aux.view
        fptr = 0
        remote = aux.peername
        if aux.freeptr > 0:
            fptr = aux.freeptr
            aux.freeptr = 0
            rbuf = view[fptr:]
        else:
            rbuf = buf

        rlen = conn.recv_into(rbuf)
        if rlen == 0:
            self._log.debug("Peer disconnected: %s.", remote)
            self._cleanup(conn, remote)
            return

        self._log.debug("%d/%d bytes received from %s.", rlen, len(rbuf), remote)
        datalen = fptr + rlen
        fptr = aux.lastptr
        aux.lastptr = 0
        cnt = 0
        while fptr < (datalen - HEADER_SIZE):
            pstart = fptr + HEADER_SIZE
            psize = int.from_bytes(view[fptr:pstart], BYTEORDER)
            pend = pstart + psize
            if psize > 0:
                if pend <= datalen:
                    chunk = bytes(view[pstart:pend])
                    self.queue.append((self.__class__, chunk, remote))
                    cnt += 1
                else:
                    break
            else:
                self._log.debug("Invalid message header: %d!", psize)
            fptr = pend
        self._log.debug("%d message(s) received.", cnt)
        if fptr != datalen:
            leftover = datalen - fptr
            self._log.debug("%d bytes leftover.", leftover)
            if fptr > len(buf) / 2:
                buf[:leftover] = buf[fptr:datalen]
                aux.freeptr = leftover
                self._log.debug("Leftover bytes moved to buffer start.")
            else:
                aux.lastptr = fptr
                aux.freeptr = datalen




# -*- generated by 1.0.6 -*-
import da
PatternExpr_260 = da.pat.TuplePattern([da.pat.ConstantPattern('Done'), da.pat.BoundPattern('_BoundPattern263_')])
PatternExpr_282 = da.pat.TuplePattern([da.pat.ConstantPattern('Request'), da.pat.FreePattern('timestamp')])
PatternExpr_289 = da.pat.FreePattern('source')
PatternExpr_317 = da.pat.TuplePattern([da.pat.ConstantPattern('Reply'), da.pat.FreePattern('c1')])
PatternExpr_324 = da.pat.FreePattern('source')
PatternExpr_266 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('Done'), da.pat.BoundPattern('_BoundPattern276_')])])
_config_object = {}
import sys

class P(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._PReceivedEvent_0 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_0', PatternExpr_260, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_1', PatternExpr_282, sources=[PatternExpr_289], destinations=None, timestamps=None, record_history=None, handlers=[self._P_handler_281]), da.pat.EventPattern(da.pat.ReceivedEvent, '_PReceivedEvent_2', PatternExpr_317, sources=[PatternExpr_324], destinations=None, timestamps=None, record_history=None, handlers=[self._P_handler_316])])

    def setup(self, ps, nrounds, **rest_402):
        super().setup(ps=ps, nrounds=nrounds, **rest_402)
        self._state.ps = ps
        self._state.nrounds = nrounds
        self._state.reqc = None
        self._state.waiting = set()
        self._state.replied = set()

    def run(self):

        def anounce():
            self.output('In cs!')
        for i in range(self._state.nrounds):
            self.cs(anounce)
        self.send(('Done', self._id), to=self._state.ps)
        super()._label('_st_label_251', block=False)
        p = None

        def UniversalOpExpr_252():
            nonlocal p
            for p in self._state.ps:
                if (not PatternExpr_266.match_iter(self._PReceivedEvent_0, _BoundPattern276_=p, SELF_ID=self._id)):
                    return False
            return True
        _st_label_251 = 0
        while (_st_label_251 == 0):
            _st_label_251 += 1
            if UniversalOpExpr_252():
                _st_label_251 += 1
            else:
                super()._label('_st_label_251', block=True)
                _st_label_251 -= 1
        self.output('Terminating..')

    def cs(self, task):
        super()._label('start', block=False)
        self._state.reqc = self.logical_clock()
        self.send(('Request', self._state.reqc), to=self._state.ps)
        super()._label('_st_label_200', block=False)
        _st_label_200 = 0
        while (_st_label_200 == 0):
            _st_label_200 += 1
            if (len(self._state.replied) == len(self._state.ps)):
                _st_label_200 += 1
            else:
                super()._label('_st_label_200', block=True)
                _st_label_200 -= 1
        task()
        super()._label('release', block=False)
        self._state.reqc = None
        self.send(('Reply', self.logical_clock()), to=self._state.waiting)
        super()._label('end', block=False)
        self._state.waiting = set()
        self._state.replied = set()

    def _P_handler_281(self, timestamp, source):
        if ((self._state.reqc == None) or ((timestamp, source) < (self._state.reqc, self._id))):
            self.send(('Reply', self.logical_clock()), to=source)
        else:
            self._state.waiting.add(source)
    _P_handler_281._labels = None
    _P_handler_281._notlabels = None

    def _P_handler_316(self, c1, source):
        if ((not (self._state.reqc is None)) and (c1 > self._state.reqc)):
            self._state.replied.add(source)
    _P_handler_316._labels = None
    _P_handler_316._notlabels = None

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])
    _config_object = {'clock': 'Lamport'}

    def run(self):
        nprocs = (int(sys.argv[1]) if (len(sys.argv) > 1) else 10)
        nrounds = (int(sys.argv[2]) if (len(sys.argv) > 2) else 1)
        ps = self.new(P, num=nprocs)
        for p in ps:
            self._setup({p}, ((ps - {p}), nrounds))
        self._start(ps)
